﻿using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using SlimDX;

using DX10 = SlimDX.Direct3D10;
using DXGI = SlimDX.DXGI;
using D3DCompiler = SlimDX.D3DCompiler;

namespace TutorialSeries.DirectX10.Chapter6
{
    public class Stars
    {
        private const int STARS_COUNT = 50;

        private List<StarInstance> m_stars;
        private Random m_randomizer;
        private DX10.Buffer m_vertexBuffer;
        private DX10.InputLayout m_vertexLayout;
        private DX10.Effect m_effect;
        private DX10.EffectTechnique m_effectTechnique;
        private DX10.EffectPass m_effectPass;
        private DX10.EffectMatrixVariable m_transformVariable;
        private DX10.EffectVectorVariable m_colorVariable;
        private DX10.Texture2D m_texture;
        private DX10.EffectResourceVariable m_resourceVariable;
        private DX10.ShaderResourceView m_textureView;
        private DX10.BlendState m_blendState;

        public void LoadResources(DX10.Device device)
        {
            Vertex[] vertices = new Vertex[]
            {
                new Vertex(new Vector3(-1f, -1f, 0f), new Vector2(0, 0)),
                new Vertex(new Vector3(-1f, 1f, 0f), new Vector2(0, 1f)),
                new Vertex(new Vector3(1f, 1f, 0f), new Vector2(1f, 1f)),
                new Vertex(new Vector3(1f, 1f, 0f), new Vector2(1f, 1f)),
                new Vertex(new Vector3(1f, -1f, 0f), new Vector2(1f, 0)),
                new Vertex(new Vector3(-1f, -1f, 0f), new Vector2(0, 0))
            };

            DataStream outStream = new DataStream(6 * Marshal.SizeOf(typeof(Vertex)), true, true);
            outStream.WriteRange(vertices);
            outStream.Position = 0;

            DX10.BufferDescription bufferDescription = new DX10.BufferDescription();
            bufferDescription.BindFlags = DX10.BindFlags.VertexBuffer;
            bufferDescription.CpuAccessFlags = DX10.CpuAccessFlags.None;
            bufferDescription.OptionFlags = DX10.ResourceOptionFlags.None;
            bufferDescription.SizeInBytes = 6 * Marshal.SizeOf(typeof(Vertex));
            bufferDescription.Usage = DX10.ResourceUsage.Default;

            m_vertexBuffer = new DX10.Buffer(device, outStream, bufferDescription);
            outStream.Close();

            m_effect = DX10.Effect.FromFile(
                device,
                "Shaders/SimpleRendering.fx", "fx_4_0",
                D3DCompiler.ShaderFlags.None, D3DCompiler.EffectFlags.None,
                null, null, null);
            m_effectTechnique = m_effect.GetTechniqueByIndex(0);
            m_effectPass = m_effectTechnique.GetPassByIndex(0);
            m_transformVariable = m_effect.GetVariableByName("WorldViewProj").AsMatrix();
            m_resourceVariable = m_effect.GetVariableByName("Texture").AsResource();
            m_colorVariable = m_effect.GetVariableByName("StarColor").AsVector();

            DX10.InputElement[] inputElements = new DX10.InputElement[]
            {
                new DX10.InputElement("POSITION",0,DXGI.Format.R32G32B32_Float,0,0),
                new DX10.InputElement("TEXCOORD",0,DXGI.Format.R32G32_Float,12,0)
            };
            m_vertexLayout = new DX10.InputLayout(device, inputElements, m_effectPass.Description.Signature);

            m_texture = DX10.Texture2D.FromFile(device, "Textures/Star.bmp");
            m_textureView = new DX10.ShaderResourceView(device, m_texture);

            m_stars = new List<StarInstance>();
            m_randomizer =new Random(Environment.TickCount);
            for (int loop = 1; loop <= STARS_COUNT; loop++)
            {
                StarInstance newInstance = new StarInstance();
                newInstance.Angle = 0f;
                newInstance.Distance = ((float)loop / (float)STARS_COUNT) * 5f;
                newInstance.Color = new Vector4(
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                    1f);
                m_stars.Add(newInstance);
            }

            DX10.BlendStateDescription blendDesc = new DX10.BlendStateDescription();
            blendDesc.BlendOperation = DX10.BlendOperation.Add;
            blendDesc.AlphaBlendOperation = DX10.BlendOperation.Add;
            blendDesc.SourceAlphaBlend = DX10.BlendOption.Zero;
            blendDesc.DestinationAlphaBlend = DX10.BlendOption.Zero;
            blendDesc.SourceBlend = DX10.BlendOption.SourceColor;
            blendDesc.DestinationBlend = DX10.BlendOption.One;
            blendDesc.IsAlphaToCoverageEnabled = false;
            blendDesc.SetWriteMask(0, DX10.ColorWriteMaskFlags.All);
            blendDesc.SetBlendEnable(0, true);
            m_blendState = DX10.BlendState.FromDescription(device, blendDesc);
        }

        public void Render(DX10.Device device, Matrix world, Matrix viewProj)
        {
            device.InputAssembler.SetInputLayout(m_vertexLayout);
            device.InputAssembler.SetPrimitiveTopology(DX10.PrimitiveTopology.TriangleList);
            device.InputAssembler.SetVertexBuffers(0, new DX10.VertexBufferBinding(m_vertexBuffer, Marshal.SizeOf(typeof(Vertex)), 0));

            m_resourceVariable.SetResource(m_textureView);
            device.OutputMerger.BlendState = m_blendState;

            for(int loop=0 ; loop<STARS_COUNT; loop++)
            {
                StarInstance actStar = m_stars[loop];

                m_colorVariable.Set(actStar.Color);
                m_transformVariable.SetMatrix((world * Matrix.Translation(actStar.Distance, 0f, 0f) * Matrix.RotationZ(actStar.Angle)) * viewProj);

                m_effectPass.Apply();
                device.Draw(6, 0);

                actStar.Angle += ((float)loop / (float)STARS_COUNT * 2f % 1f) / 57f;
                if (actStar.Angle > (float)Math.PI * 2f) { actStar.Angle = actStar.Angle - (float)Math.PI * 2f; }

                actStar.Distance = actStar.Distance - 0.01f;
                if (actStar.Distance < 0f)
                {
                    actStar.Distance = 5.0f;
                    actStar.Color = new Vector4(
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                        m_randomizer.Next(128, 255) / 255f,
                        1f);
                }
            }
        }

        //*********************************************************************
        //*********************************************************************
        //*********************************************************************
        class StarInstance
        {
            public Vector4 Color;
            public float Angle;
            public float Distance;
        }
    }
}
